@@ -34,9 +34,17 @@
  * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
  * 02110-1301, USA.
  *
+ *
+ * Changes:
+ *     Hirotaka Sasaki <hiro1967@mti.biglobe.ne.jp>:
+ *         - Supported Linux-2.6.
+ *         - Rewrote to remove tcps module.
+ *         - Rewrote to work without kernel patch.
+ *
  **********************************************************************/
 
 #include <linux/config.h>
+#include <linux/version.h>
 #include <linux/kernel.h>
 #include <linux/module.h>
 #include <linux/init.h>
@@ -63,11 +71,8 @@
 
 #include <asm/uaccess.h>
 
-#ifdef TCPS_INTEGRATED_SOURCE
-#include <net/ipv4/tcps/tcps.h>
-#else
 #include "tcps.h"
-#endif
+#include "tcps_compat.h"
 
 union tcps_tphdr {
 	unsigned char *raw;
@@ -174,16 +179,59 @@ tcps_fast_parse_timestamps(struct tcphdr *th, __u32 **ts_p)
 	return tcps_parse_timestamps(th, ts_p);
 }
 
+static void
+tcps_timestamps_to_nop(struct tcphdr *th)
+{
+	unsigned char *p;
+	int len = (th->doff * 4) - sizeof(struct tcphdr);
+
+	p = (unsigned char *)(th + 1);
+
+	while(len > 0) {
+	  	int opt = *p;
+		int olen;
+
+		switch (opt) {
+		case TCPOPT_EOL:
+			return;
+
+		case TCPOPT_NOP:
+			p++;
+			len--;
+			continue;
+		}
+
+		olen = *(p+1);
+		if (olen < 2 || olen > len) {
+			return;
+		}
+
+		switch (opt) {
+		case TCPOPT_TIMESTAMP:
+			if (olen == TCPOLEN_TIMESTAMP) {
+				memset(p, TCPOPT_NOP, TCPOLEN_TIMESTAMP);
+			}
+			return;
+	  	}
+		p += olen;
+		len -= olen;
+	}
+	return;
+}
+
 void
 tcps_reset_sock(struct sock *sk)
 {
-	TCPS_DBG("resetting socket...\n");
-	sk->err = ECONNRESET;
-	if (!sk->dead) {
-		sk->error_report(sk);
-	}
+	TCPS_DBG("tcps_reset_sock: sk=%p\n", sk);
 
-	tcp_done(sk);
+	if (TCPS_SK_STATE(sk) != TCP_CLOSE) {
+		TCPS_SK_ERR(sk) = ECONNRESET;
+		if (!TCPS_SK_FL_DEAD(sk)) {
+			TCPS_SK_ERR_REPORT(sk);
+		}
+
+		tcp_done(sk);
+	}
 }
 
 static struct sock *
@@ -199,7 +247,7 @@ tcps_sock_get(int fd, int *err)
 	}
 
 	inode = file->f_dentry->d_inode;
-	if (!inode->i_sock || !(sock = &inode->u.socket_i)) {
+	if (!inode->i_sock || !(sock = SOCKET_I(inode))) {
 		*err = -ENOTSOCK;
 		fput(file);
 		return NULL;
@@ -221,7 +269,7 @@ tcps_conn_init_tcpopt(struct tcps_conn *tc, struct sock *sk, int which)
 	struct tcp_opt *tp;
 	struct tcps_conn_tcpopt *tcopt;
 
-	tp = &sk->tp_pinfo.af_tcp;
+	tp = tcp_sk(sk);
 	tcopt = &tc->tcpopt[which];
 
 	tcopt->sseq = tp->snd_nxt;
@@ -242,33 +290,45 @@ tcps_splice(struct sock *sk1, struct sock *sk2)
 {
 	struct tcp_opt *tp1, *tp2;
 	struct tcps_conn *tc;
+	struct tcps_info *tcpsi;
 
-	if (sk1->protocol != sk2->protocol || sk1->protocol != IPPROTO_TCP) {
+	if (TCPS_SK_PROTOCOL(sk1) != TCPS_SK_PROTOCOL(sk2)
+	    || TCPS_SK_PROTOCOL(sk1) != IPPROTO_TCP) {
 		TCPS_DBG("tcps_splice: protocol mismatch\n");
 		return -EINVAL;
 	}
 
-	if ((sk2->tcpsinfo.flags & TCPS_F_SOURCE_FAKE) != TCPS_F_SOURCE_FAKE) {
-		TCPS_DBG("tcps_splice: sk2 flag error\n");
+	tcpsi = tcps_info_get(sk2);
+    if (!tcpsi) {
+		TCPS_DBG("tcps_splice: not found info: sk2=%p \n", sk2);
 		return -EINVAL;
-	}
+	} else if (!(tcpsi->flags & TCPS_INFO_F_SOURCE_FAKE)) {
+		TCPS_DBG("tcps_splice: info flags error: sk2=%p tcpsi=%p\n",
+			sk2, tcpsi);
+		tcps_info_put(sk2);
+		return -EINVAL;
+    }
+	spin_lock_bh(&tcpsi->lock);
+	tcpsi->flags |= TCPS_INFO_F_SPLICED;
+	spin_unlock_bh(&tcpsi->lock);
+	tcps_info_put(sk2);
 
-	tp1 = &sk1->tp_pinfo.af_tcp;
-	tp2 = &sk2->tp_pinfo.af_tcp;
-	
-	tc = tcps_conn_preroute_get(sk1->daddr, sk1->dport,
-				    sk2->daddr, sk2->dport);
+	tp1 = tcp_sk(sk1);
+	tp2 = tcp_sk(sk2);
+
+	tc = tcps_conn_preroute_get(TCPS_SK_DADDR(sk1), TCPS_SK_DPORT(sk1),
+				    TCPS_SK_DADDR(sk2), TCPS_SK_DPORT(sk2));
 	if (tc == NULL) {
 		TCPS_DBG("tcps_splice: connection entry not found\n");
 		TCPS_DBG("sk1=%08x:%d sk2=%08x:%d\n",
-			 htonl(sk1->daddr), htons(sk1->dport),
-			 htonl(sk2->daddr), htons(sk2->dport));
+			 htonl(TCPS_SK_DADDR(sk1)), htons(TCPS_SK_DPORT(sk1)),
+			 htonl(TCPS_SK_DADDR(sk2)), htons(TCPS_SK_DPORT(sk2)));
 		return -EINVAL;
 	}
 
 	/* now we can splice sockets */
-	tc->vaddr = sk1->rcv_saddr;
-	tc->vport = sk1->sport;
+	tc->vaddr = TCPS_SK_RCVADDR(sk1);
+	tc->vport = TCPS_SK_SPORT(sk1);
 
 	/* store tcp connection parameters */
 	tcps_conn_init_tcpopt(tc, sk1, TCPS_CONN_TCPOPT_CL);
@@ -293,6 +353,7 @@ tcps_setsockopt(struct sock *sk, int cmd, void *user, unsigned int len)
 {
 	struct sockaddr_in sin;
 	struct tcps_tcpopt tcpopt;
+	struct tcps_info *tcpsi;
 	struct sock *sk2;
 	int spfd;
 	int ret = 0;
@@ -304,15 +365,28 @@ tcps_setsockopt(struct sock *sk, int cmd, void *user, unsigned int len)
 			ret = -EINVAL;
 			break;
 		}
-		
+
 		if (copy_from_user(&sin, user, len) != 0) {
 			ret = -EFAULT;
 			break;
 		}
 
-		sk->tcpsinfo.flags = TCPS_F_SOURCE_FAKE;
-		sk->tcpsinfo.saddr = sin.sin_addr.s_addr;
-		sk->tcpsinfo.sport = sin.sin_port;
+		tcpsi= tcps_info_get(sk);
+		if (!tcpsi) {
+			tcpsi = tcps_info_new(sk);
+			if (IS_ERR(tcpsi)) {
+				ret = PTR_ERR(tcpsi);
+				break;
+			}
+		}
+
+		spin_lock_bh(&tcpsi->lock);
+		tcpsi->saddr = sin.sin_addr.s_addr;
+		tcpsi->sport = sin.sin_port;
+		tcpsi->flags |= TCPS_INFO_F_SOURCE_FAKE;
+		spin_unlock_bh(&tcpsi->lock);
+
+		tcps_info_put(sk);
 		break;
 
 	case TCPS_SO_SPLICE:
@@ -330,18 +404,21 @@ tcps_setsockopt(struct sock *sk, int cmd, void *user, unsigned int len)
 
 		sk2 = tcps_sock_get(spfd, &ret);
 		if (sk2 == NULL) {
+			ret = -EINVAL;
 			break;
 		}
 
-		if (sk2->tcpsinfo.flags == 0) {
-			ret = -EINVAL;
+		sock_hold(sk);
+		ret = tcps_splice(sk, sk2);
+		if (ret) {
+			sock_put(sk);
 			sock_put(sk2);
 			break;
 		}
 
-		TCPS_DBG("tcps_setsockopt(SPLICE): splicing...\n");
-		sock_hold(sk);
-		ret = tcps_splice(sk, sk2);
+		TCPS_DBG("tcps_setsockopt(SPLICE): splicing succeed : sk=%p sk2=%p\n",
+			 sk, sk2);
+
 		break;
 
 	case TCPS_SO_SET_TCPOPT:
@@ -357,7 +434,20 @@ tcps_setsockopt(struct sock *sk, int cmd, void *user, unsigned int len)
 			break;
 		}
 
-		sk->tcpsinfo.suppress_tstamp = !tcpopt.tstamp_ok;
+		tcpsi= tcps_info_get(sk);
+		if (!tcpsi) {
+			tcpsi = tcps_info_new(sk);
+			if (IS_ERR(tcpsi)) {
+				ret = PTR_ERR(tcpsi);
+				break;
+			}
+		}
+
+		spin_lock_bh(&tcpsi->lock);
+		tcpsi->suppress_tstamp = !tcpopt.tstamp_ok;
+		spin_unlock_bh(&tcpsi->lock);
+
+		tcps_info_put(sk);
 		break;
 
 	default:
@@ -383,7 +473,7 @@ tcps_getsockopt(struct sock *sk, int cmd, void *user, int *len)
 			break;
 		}
 
-		tp = &sk->tp_pinfo.af_tcp;
+		tp = tcp_sk(sk);
 		tcpopt.tstamp_ok = tp->tstamp_ok;
 
 		if (copy_to_user(user, &tcpopt, sizeof(tcpopt))) {
@@ -401,9 +491,14 @@ tcps_getsockopt(struct sock *sk, int cmd, void *user, int *len)
 }
 
 static struct nf_sockopt_ops tcps_sockopts = {
-	{NULL, NULL}, PF_INET,
-	TCPS_SO_BASE, TCPS_SO_SET_MAX + 1, tcps_setsockopt,
-	TCPS_SO_BASE, TCPS_SO_GET_MAX + 1, tcps_getsockopt
+	.list       = {NULL, NULL},
+	.pf         = PF_INET,
+	.set_optmin = TCPS_SO_BASE,
+	.set_optmax = TCPS_SO_SET_MAX + 1,
+	.set        = tcps_setsockopt,
+	.get_optmin = TCPS_SO_BASE,
+	.get_optmax = TCPS_SO_GET_MAX + 1,
+	.get        = tcps_getsockopt
 };
 
 static int
@@ -437,9 +532,11 @@ tcps_out(unsigned int hooknum,
 	struct rtable *rt;
 	struct sock *sk = skb->sk;
 	struct iphdr *iph;
+	struct tcps_info *tcpsi = NULL;
 	union tcps_tphdr h;
 	struct tcps_conn *tc = NULL;
 	struct tcps_conn_tcpopt *ctcopt, *rtcopt;
+	u32 tcpsi_state = 0;
 	int ihl, datalen;
 	int ret;
 
@@ -449,7 +546,13 @@ tcps_out(unsigned int hooknum,
 	}
 
 	sock_hold(sk);
-	if (!sk->tcpsinfo.flags) {
+
+	tcpsi = tcps_info_get(sk);
+    if (!tcpsi) {
+		ret = NF_ACCEPT;
+		goto out;
+	} else if (!(tcpsi->flags & TCPS_INFO_F_SOURCE_FAKE)) {
+		tcps_info_put(sk);
 		ret = NF_ACCEPT;
 		goto out;
 	}
@@ -464,17 +567,13 @@ tcps_out(unsigned int hooknum,
 		goto out;
 	}
 
-	if (ip_route_output(&rt, iph->daddr, 0, RT_TOS(iph->tos), 0)) {
-		TCPS_DBG("ip_route_output failed\n");
+	if (tcps_ip_route_output(&rt, iph->daddr, 0, RT_TOS(iph->tos), 0)) {
+		TCPS_DBG("tcps_ip_route_output failed\n");
 		ret = NF_DROP;
 		goto out;
 	}
 
-	if (h.th->fin) {
-		TCPS_DBG("tcps_out: FIN\n");
-	}
-
-	tc = tcps_conn_out_rs_get(sk->tcpsinfo.saddr, sk->tcpsinfo.sport,
+	tc = tcps_conn_out_rs_get(tcpsi->saddr, tcpsi->sport,
 				  iph->saddr, h.th->source);
 				  
 	if (tc == NULL) {
@@ -482,19 +581,39 @@ tcps_out(unsigned int hooknum,
 			ret = NF_ACCEPT;
 			goto out;
 		}
+
 		tc = tcps_conn_new(iph->saddr, h.th->source,
 				   iph->daddr, h.th->dest, 
-				   sk->tcpsinfo.saddr, sk->tcpsinfo.sport);
+				   tcpsi->saddr, tcpsi->sport);
 		if (tc == NULL) {
 			ret = NF_DROP;
 			goto out;
 		}
-	} else {
-		TCPS_DBG(KERN_INFO "tcps_out: found entry %x:%d -> %x:%d\n",
+		TCPS_DBG("tcps_out: new entry %x:%d -> %x:%d\n",
 			 ntohl(tc->caddr), ntohs(tc->cport), 
 			 ntohl(tc->raddr), ntohs(tc->rport));
+
+		tcpsi_state = TCPS_INFO_S_SYN;
+	} else {
+		TCPS_DBG("tcps_out: found entry %x:%d -> %x:%d\n",
+			 ntohl(tc->caddr), ntohs(tc->cport), 
+			 ntohl(tc->raddr), ntohs(tc->rport));
+
+		if (h.th->syn) {
+			tcpsi_state = TCPS_INFO_S_SYN;
+		} else if (h.th->fin) {
+			tcpsi_state = TCPS_INFO_S_FIN;
+		} else if (h.th->rst) {
+			tcpsi_state = TCPS_INFO_S_RST;
+		} else {
+			tcpsi_state = TCPS_INFO_S_SENT;
+		}
 	}
 
+	spin_lock_bh(&tcpsi->lock);
+	tcps_info_set_state(tcpsi, tcpsi->state|tcpsi_state);
+	spin_unlock_bh(&tcpsi->lock);
+
 	spin_lock_bh(&tc->lock);
 	ctcopt = &tc->tcpopt[TCPS_CONN_TCPOPT_CL];
 	rtcopt = &tc->tcpopt[TCPS_CONN_TCPOPT_RS];
@@ -504,20 +623,27 @@ tcps_out(unsigned int hooknum,
 		goto out;
 	}
 
-	if ((tc->state & TCPS_CONN_S_SPLICED) == TCPS_CONN_S_SPLICED) {
-		int plen;
-		plen = (int)(ntohl(h.th->seq) - rtcopt->sseq) + 
-			datalen - (h.th->doff << 2);
-		TCPS_DBG("tcps_out: seq=%u nseq=%u\n",
-			 rtcopt->csseq, rtcopt->csseq + plen);
+	if (h.th->syn && tcpsi->suppress_tstamp) {
+		tcps_timestamps_to_nop(h.th);
+	}
 
-		h.th->seq = htonl(rtcopt->csseq);
-		h.th->ack_seq = htonl(rtcopt->lack);
-		if (plen == 0 && h.th->fin) {
-			plen++;
+	if ((tc->state & TCPS_CONN_S_SPLICED) == TCPS_CONN_S_SPLICED) {
+
+		if (!before(ntohl(h.th->seq), rtcopt->csseq)) {
+			int plen;
+			plen = (int)(ntohl(h.th->seq) - rtcopt->sseq) + 
+			    datalen - (h.th->doff << 2);
+			TCPS_DBG("tcps_out: seq=%u nseq=%u\n",
+			    rtcopt->csseq, rtcopt->csseq + plen);
+
+			h.th->seq = htonl(rtcopt->csseq);
+			h.th->ack_seq = htonl(rtcopt->lack);
+			if (plen == 0 && h.th->fin) {
+				plen++;
+			}
+			rtcopt->csseq += plen;
+			rtcopt->sseq += plen;
 		}
-		rtcopt->csseq += plen;
-		rtcopt->sseq += plen;
 	}
 
 	/* virtualserver->realserver. mangle the source to client address. */
@@ -526,24 +652,32 @@ tcps_out(unsigned int hooknum,
 
 	spin_unlock_bh(&tc->lock);
 	tcps_conn_put(tc);
+	tcps_info_put(sk);
 	sock_put(sk);
 
 	h.th->check = 0;
  	h.th->check = csum_tcpudp_magic(iph->saddr, iph->daddr,
 					datalen, iph->protocol,
 					csum_partial(h.raw, datalen, 0));
-	if (skb->ip_summed == CHECKSUM_HW)
-		skb->ip_summed = CHECKSUM_NONE;
+
+	skb->ip_summed = CHECKSUM_UNNECESSARY;
+	ip_send_check(iph);
 
 	(*okfn)(skb);
 	return NF_STOLEN;
 
  out:
+	if (tcpsi != NULL) {
+		tcps_info_put(sk);
+	}
+
 	if (tc != NULL) {
 		spin_unlock_bh(&tc->lock);
 		tcps_conn_put(tc);
 	}
+
 	sock_put(sk);
+
 	return ret;
 }
 
@@ -563,6 +697,7 @@ tcps_in(unsigned int hooknum,
 	u32 seq, ack;
 	u32 tsval, tsecho, *ts;
 	u32 state;
+	u32 tcpsi_state = 0;
 	int ihl, datalen;
 	int have_ts = 0;
 
@@ -588,13 +723,18 @@ tcps_in(unsigned int hooknum,
 		/* The client is re-using port number... */
 		spin_unlock_bh(&tc->lock);
 		tcps_conn_expire_now(tc);
+		__tcps_conn_put(tc);
 		return NF_ACCEPT;
 	}
 
 	if (h.th->fin) {
-		state = tc->state | TCPS_CONN_S_CL_FIN;
+		state = TCPS_CONN_S_CL_FIN;
+		tcpsi_state = TCPS_INFO_S_FIN;
+	} else if (h.th->rst) {
+		state = TCPS_CONN_S_CL_RST;
+		tcpsi_state = TCPS_INFO_S_RST;
 	} else {
-		state = tc->state | TCPS_CONN_S_CL_SENT;
+		state = TCPS_CONN_S_CL_SENT;
 	}
 
 	ctcopt = &tc->tcpopt[TCPS_CONN_TCPOPT_CL];
@@ -604,8 +744,8 @@ tcps_in(unsigned int hooknum,
 		 ntohl(tc->caddr), ntohs(tc->cport), 
 		 ntohl(tc->raddr), ntohs(tc->rport));
 
-	if (ip_route_output(&rt, tc->raddr, 0, RT_TOS(iph->tos), 0)) {
-		TCPS_DBG("ip_route_output failed\n");
+	if (tcps_ip_route_output(&rt, tc->raddr, 0, RT_TOS(iph->tos), 0)) {
+		TCPS_DBG("tcps_ip_route_output failed\n");
 		spin_unlock_bh(&tc->lock);
 		tcps_conn_put(tc);
 		return NF_DROP;
@@ -647,13 +787,32 @@ tcps_in(unsigned int hooknum,
 		rtcopt->lack = ack;
 	}
 
-	if (h.th->fin && tc->csk != NULL) {
-		tcps_reset_sock(tc->csk);
-		sock_put(tc->csk);
-		tc->csk = NULL;
+	if (h.th->fin | h.th->rst) {
+		if (tc->csk != NULL) {
+			tcps_reset_sock(tc->csk);
+			sock_put(tc->csk);
+			tc->csk = NULL;
+		}
+
+		if (tc->rsk != NULL) {
+			struct tcps_info *tcpsi;
+
+			if ((tcpsi = tcps_info_get(tc->rsk)) != NULL) {
+				spin_lock_bh(&tcpsi->lock);
+				tcps_info_set_state(tcpsi, tcpsi->state|tcpsi_state);
+				spin_unlock_bh(&tcpsi->lock);
+				tcps_info_put(tcpsi->sk);
+			}
+
+			if (h.th->rst) {
+				tcps_reset_sock(tc->rsk);
+				sock_put(tc->rsk);
+				tc->rsk = NULL;
+			}
+		}
 	}
 
-	tcps_conn_set_state(tc, state);
+	tcps_conn_set_state(tc, tc->state|state);
 	spin_unlock_bh(&tc->lock);
 	tcps_conn_put(tc);
 
@@ -665,7 +824,7 @@ tcps_in(unsigned int hooknum,
 	skb->ip_summed = CHECKSUM_UNNECESSARY;
 	ip_send_check(iph);
 
-	ip_send(skb);
+	tcps_ip_send(skb);
 	return NF_STOLEN;
 }
 
@@ -686,8 +845,10 @@ tcps_preroute(unsigned int hooknum,
 	u32 tsval, tsecho, *ts;
 	u32 seq, ack;
 	u32 state;
+	u32 tcpsi_state = 0;
 	int for_local = 0;
 	int have_ts = 0;
+	int is_spliced, is_already_rst;
 	int (*sndfn)(struct sk_buff *);
 
 	iph = skb->nh.iph;
@@ -713,12 +874,19 @@ tcps_preroute(unsigned int hooknum,
 	ack = ntohl(h.th->ack_seq);
 
 	if (h.th->fin) {
-		state = tc->state | TCPS_CONN_S_RS_FIN;
+		state = TCPS_CONN_S_RS_FIN;
+		tcpsi_state = TCPS_INFO_S_FIN;
+	} else if (h.th->rst) {
+		state = TCPS_CONN_S_RS_RST;
+		tcpsi_state = TCPS_INFO_S_RST;
+	} else if (h.th->syn) {
+		state = TCPS_CONN_S_RS_SYN;
 	} else {
-		state = tc->state | TCPS_CONN_S_RS_SENT;
+		state = TCPS_CONN_S_RS_SENT;
 	}
 
-	if ((tc->state & TCPS_CONN_S_SPLICED) != TCPS_CONN_S_SPLICED) {
+	is_spliced = tc->state & TCPS_CONN_S_SPLICED;
+	if (!is_spliced) {
 		for_local = 1;
 	} else {
 		if (h.th->ack && datalen == h.th->doff * 4 &&
@@ -731,7 +899,9 @@ tcps_preroute(unsigned int hooknum,
 
 	if (for_local) {
 		/* route to local socket w/ destination-address mangling */
-		if (tc->rsk == NULL && (tc->state & TCPS_CONN_S_RS_FIN)) {
+		is_already_rst = tc->state
+		    & (TCPS_CONN_S_RS_FIN|TCPS_CONN_S_RS_RST|TCPS_CONN_S_CL_RST);
+		if (is_spliced && is_already_rst && tc->rsk == NULL) {
 			/* Already reset, drop it. */
 			TCPS_DBG("socket already reset, drop it\n");
 			spin_unlock_bh(&tc->lock);
@@ -782,8 +952,8 @@ tcps_preroute(unsigned int hooknum,
 		TCPS_DBG("tcps_preroute: seq=%u ack=%u tsval=%u tsecr=%u\n",
 			 seq, ack, tsval, tsecho);
 
-		if (ip_route_output(&rt, iph->daddr, 0, RT_TOS(iph->tos), 0)) {
-			TCPS_DBG("ip_route_output failed\n");
+		if (tcps_ip_route_output(&rt, iph->daddr, 0, RT_TOS(iph->tos), 0)) {
+			TCPS_DBG("tcps_ip_route_output failed\n");
 			spin_unlock_bh(&tc->lock);
 			tcps_conn_put(tc);
 			return NF_DROP;
@@ -791,16 +961,34 @@ tcps_preroute(unsigned int hooknum,
 		
 		dst_release(skb->dst);
 		skb->dst = &rt->u.dst;
-		sndfn = ip_send;
+		sndfn = tcps_ip_send;
 	}
 
-	if (h.th->fin && tc->rsk != NULL) {
-		tcps_reset_sock(tc->rsk);
-		sock_put(tc->rsk);
-		tc->rsk = NULL;
+	if (h.th->fin | h.th->rst) {
+
+		if (tc->rsk != NULL) {
+			struct tcps_info *tcpsi;
+
+			if ((tcpsi = tcps_info_get(tc->rsk)) != NULL) {
+				spin_lock_bh(&tcpsi->lock);
+				tcps_info_set_state(tcpsi, tcpsi->state|tcpsi_state);
+				spin_unlock_bh(&tcpsi->lock);
+				tcps_info_put(tcpsi->sk);
+			}
+
+			tcps_reset_sock(tc->rsk);
+			sock_put(tc->rsk);
+			tc->rsk = NULL;
+		}
+
+		if (h.th->rst && tc->csk != NULL) {
+			tcps_reset_sock(tc->csk);
+			sock_put(tc->csk);
+			tc->csk = NULL;
+		}
 	}
 
-	tcps_conn_set_state(tc, state);
+	tcps_conn_set_state(tc, tc->state|state);
 	spin_unlock_bh(&tc->lock);
 	tcps_conn_put(tc);
 
@@ -818,18 +1006,36 @@ tcps_preroute(unsigned int hooknum,
 }
 
 static struct nf_hook_ops tcps_out_ops = {
-	{ NULL, NULL },
-	tcps_out, PF_INET, NF_IP_LOCAL_OUT, 100
+	.list     = { NULL, NULL },
+	.hook     = tcps_out,
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
+	.owner    = THIS_MODULE,
+#endif
+	.pf       = PF_INET,
+	.hooknum  = NF_IP_LOCAL_OUT,
+	.priority = 100
 };
 
 static struct nf_hook_ops tcps_in_ops = {
-	{ NULL, NULL },
-	tcps_in, PF_INET, NF_IP_LOCAL_IN, 100
+	.list     = { NULL, NULL },
+	.hook     = tcps_in,
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
+	.owner    = THIS_MODULE,
+#endif
+	.pf       = PF_INET,
+	.hooknum  = NF_IP_LOCAL_IN,
+	.priority = 100
 };
 
 static struct nf_hook_ops tcps_preroute_ops = {
-	{ NULL, NULL },
-	tcps_preroute, PF_INET, NF_IP_PRE_ROUTING, 100
+	.list     = { NULL, NULL },
+	.hook     = tcps_preroute,
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
+	.owner    = THIS_MODULE,
+#endif
+	.pf       = PF_INET,
+	.hooknum  = NF_IP_PRE_ROUTING,
+	.priority = 100
 };
 
 static int __init
@@ -837,21 +1043,29 @@ tcps_init(void)
 {
 	int ret;
 
-	ret = tcps_control_init();
+	ret = tcps_info_init();
 	if (ret) {
 		return ret;
 	}
 
 	ret = tcps_conn_init();
 	if (ret) {
-		tcps_control_fini();
+		tcps_info_fini();
+		return ret;
+	}
+
+	ret = tcps_control_init();
+	if (ret) {
+		tcps_conn_fini();
+		tcps_info_fini();
 		return ret;
 	}
 
 	ret = nf_register_hook(&tcps_out_ops);
 	if (ret) {
-		tcps_conn_fini();
 		tcps_control_fini();
+		tcps_conn_fini();
+		tcps_info_fini();
 		return ret;
 	}
 
@@ -859,6 +1073,8 @@ tcps_init(void)
 	if (ret) {
 		nf_unregister_hook(&tcps_out_ops);
 		tcps_control_fini();
+		tcps_conn_fini();
+		tcps_info_fini();
 		return ret;
 	}
 
@@ -866,14 +1082,13 @@ tcps_init(void)
 	if (ret) {
 		nf_unregister_hook(&tcps_in_ops);
 		nf_unregister_hook(&tcps_out_ops);
-		tcps_conn_fini();
 		tcps_control_fini();
+		tcps_conn_fini();
+		tcps_info_fini();
 		return ret;
 	}
 
-	printk(KERN_INFO "tcps module loaded.\n");
-	/* XXX  Should be fixed. Should maintain module counter correctly. */
-	MOD_INC_USE_COUNT;
+	TCPS_INF("tcps module loaded.\n");
 	return ret;
 }
 
@@ -887,10 +1102,11 @@ tcps_fini(void)
 	nf_unregister_hook(&tcps_out_ops);
 	nf_unregister_hook(&tcps_preroute_ops);
 
-	tcps_conn_fini();
 	tcps_control_fini();
+	tcps_conn_fini();
+	tcps_info_fini();
 
-	printk(KERN_INFO "tcps module unloaded.\n");
+	TCPS_INF("tcps module unloaded.\n");
 }
 
 module_exit(tcps_fini);
